{-
  PP, laboratorul 8: tipuri de date utilizator
-}

import Data.List
import TestPP

{-
  1. (2p) Numere complexe
  Se dă tipul de date Complex, reprezentând numerele complexe, i.e.
  primul câmp îl reprezintă partea reală, iar cel de-al doilea partea
  imaginară.

  Implementați următoarele operații cu numere complexe:
  - compararea la egalitate (eqC) între două numere
  - adunarea (addC) și scăderea (subC) a două numere
  - înmulțirea (mulC) și împărțirea (divC) a două numere
  - conjugata (conjC) unui număr
  - modulul (absC) unui număr

  Pentru mai multe detalii, consultați:
  https://en.wikipedia.org/wiki/Complex_number#Elementary_operations
-}
test1 :: TestPP ()
test1 = tests 1 2
        [ testCond (eqC c1 c1) "eqC" 0.5
	, testCond (((c1 `addC` c2) `subC` (C 1 1)) `eqC` (C 6 6)) "addC, subC" 0.25
	, testCond ((c1 `mulC` c2) `eqC` (C 2 23)) "mulC" 0.25
	, testCond ((c2 `divC` c1) `eqC` (C 1.6923 0.5384)) "divC" 0.25
	, testCond ((conjC c2) `eqC` (C 4 (-5))) "conjC" 0.25
	, testCond ((absC c2) `eqR` 6.4031) "absC" 0.5
	]
    where
          c1 = C 3 2
          c2 = C 4 5
	  eqR a b = abs (a - b) < 0.0001


data Complex = C
  { realC :: Double
  , imgC  :: Double
  } deriving Show

-- Hint: dat fiind că numerele complexe sunt definite pe baza a două
-- numere în virgulă mobilă, puteți implementa egalitatea aproximativă
-- dintre două numere, folosind absC și subC.
eqC :: Complex -> Complex -> Bool
eqC c1 c2 = (absC $ c1 `subC` c2) < eps
  where eps = 0.0001

addC :: Complex -> Complex -> Complex
addC (C a b) (C c d) = C (a + c) (b + d)

subC :: Complex -> Complex -> Complex
subC (C a b) (C c d) = C (a - c) (b - d)

mulC :: Complex -> Complex -> Complex
mulC (C a b) (C c d) = C (a * c - b * d) (b * c + a * d)

divC :: Complex -> Complex -> Complex
divC (C a b) (C c d) = C ((a * c + b * d) / cd2) ((b * c - a * d) / cd2)
  where cd2 = c * c + d * d

conjC :: Complex -> Complex
conjC (C a b) = C a (-b)

absC :: Complex -> Double
absC (C a b) = sqrt $ a * a + b * b



{-
 2. (4p) Liste imbricate
  Definiți un tip de date SList a care să aibă funcționalități
  asemănătoare listelor din limbajele Lisp (e.g. Scheme, Racket, Clojure),
  permițând componente la diferite niveluri de imbricare.
  Ex: Lista din Racket '(1 (3 4) (2)) să poată fi definită în Haskell
  folosind SList.
  Adițional, definiți:
  - emptySList, lista vidă
  - consElem, adaugă un element în capul unei liste
    Ex: consElem 1 '((3 4) (2)) == '(1 (3 4) (2))
  - consList, adaugă o listă (imbricată) în capul unei liste
    Ex: consList '(2 3) '(1 2) == '((2 3) 1 2)
  - headSList, ia primul element dintr-un SList
  - tailSList, ia restul SList-ului
  - deepEqual, o funcție ce verifică egalitatea a două SList
  - flatten, întoarce lista cu elementele din SList (pe același nivel)
  Notare:
  (2p) constructorii (emptySList, consElem și consList) și deepEqual
  (1p) headSList și tailSList
  (1p) flatten
-}

test2 :: TestPP ()
test2 = tests 2 4
        [ testCond (deepEqual l1 l1 && not (deepEqual l1 l2)) "simple lists" 1
	, testCond (deepEqual (consElem 2 $ l3) l2) "less simple lists" 1
	, testCond (deepEqual (headSList $ tailSList l2)
	                      (consElem 1 $ consElem 1 emptySList)) "head, tail" 1
	, testVal (flatten l2) [2, 1, 1, 3] "flatten" 1
	]
    where
          l1 = consElem 1 $ emptySList
          l2 = consElem 2 $ consList (consElem 1 $ consElem 1 emptySList) $
               consElem 3 emptySList
          l3 = consList (consElem 1 $ consElem 1 emptySList) $ consElem 3 $
               emptySList


data SList a = Atom a | List [SList a] deriving Show

emptySList :: SList a
emptySList = List []

consElem :: a -> SList a -> SList a
consElem x (Atom y)  = List $ [Atom x, Atom y]
consElem x (List xs) = List $ (Atom x : xs)

consList :: SList a -> SList a -> SList a
consList x (Atom y)  = List [x, Atom y]
consList x (List ys) = List $ x : ys

headSList :: SList a -> SList a
headSList (List x) = head x
headSList _        = error "head undefined on Atoms"

tailSList :: SList a -> SList a
tailSList (List x) = List $ tail x
tailSList _        = error "tail undefined on Atoms"

deepEqual :: Eq a => SList a -> SList a -> Bool
deepEqual (Atom x) (Atom y) = x == y
deepEqual (List a) (List b) = and $ zipWith deepEqual a b
deepEqual _ _               = False

flatten :: SList a -> [a]
flatten (Atom x)  = [x]
flatten (List xs) = concatMap flatten xs


{-
  3. (4p) Arbori binari de căutare
  Definiți un tip de date BST a pentru a implementa un arbore binar de
  căutare.  De asemenea, definiți funcții pentru a crea un arbore binar de
  căutare pornind de la o listă de elemente, căutarea unui element într-un
  arbore binar de căutare dat și o funcție care întoarce lista elementelor
  din parcurgerea în inordine a arborelui.

  Hint: Este de preferat ca arborele binar de căutare să fie echilibrat,
  lucru ușor de obținut la creare dacă lista de elemente este sortată.
-}

test3 :: TestPP ()
test3 = tests 3 4
        [ testVal (findElem bst 1) (Just 1) "findElem" 2
	, testVal (inorder bst) [1, 2, 3, 4] "inorder" 2
	]
    where
         bst = makeBST [2, 3, 1, 4]


data BST a = BSTNod a (BST a) (BST a) | BSTNil deriving Show

makeBST :: Ord a => [a] -> BST a
makeBST lst = makeBSTHelper (sort lst)

makeBSTHelper [] = BSTNil
makeBSTHelper lst = BSTNod m (makeBSTHelper st) (makeBSTHelper dr)
  where
  l = length lst
  m = lst !! (div l 2)
  st = take (div l 2) lst
  dr = drop ((div l 2) + 1) lst

findElem :: (Ord a, Eq a) => BST a -> a -> Maybe a
findElem BSTNil _ = Nothing
findElem (BSTNod x arbSt arbDr ) y
      | x == y = Just x
      | y < x = findElem arbSt y
      | otherwise = findElem arbDr y

inorder :: BST a -> [a]
inorder BSTNil                 = []
inorder (BSTNod x arbSt arbDr) = (inorder arbSt) ++ [x] ++ (inorder arbDr)



{-
  4. (BONUS, 2p) Arbori binari de căutare
  Lucrând pe același tip de date de la exercițiul anterior, definiți
  funcția subtree, care verifică dacă arborele t1 este un subarbore al
  arborelui t2.
  Ex: subtree (makeBST [1,2]) (makeBST [1,2,3,4]) == True
-}


test4 :: TestPP ()
test4 = tests 4 2
        [ testCond (subtree (makeBST [1]) t) "subtree [1]" 0.5
        , testCond (subtree (makeBST [1, 2]) t) "subtree [1, 2]" 0.5
	, testCond (subtree t t) "subtree [i, 2, 3, 4]" 0.5
	, testCond (not $ subtree (makeBST [3, 4]) t) "not $ subtree [3, 4]" 0.5
	]
    where
          t = makeBST [1, 2, 3, 4]


subtree :: (Eq a, Ord a) => BST a -> BST a -> Bool
subtree BSTNil _ = True
subtree _ BSTNil = False
subtree s@(BSTNod x1 _ _) t@(BSTNod x2 arbSt arbDr)
           | eqTree s t = True
	   | x1 < x2 = subtree s arbSt
	   | x1 > x2 = subtree s arbDr
	   | otherwise = False

eqTree :: Eq a => BST a -> BST a -> Bool
eqTree BSTNil BSTNil = True
eqTree BSTNil _ = False
eqTree _ BSTNil = False
eqTree (BSTNod x1 arbSt1 arbDr1) (BSTNod x2 arbSt2 arbDr2)
           | x1 == x2 = eqTree arbSt1 arbSt2 && eqTree arbDr1 arbDr2
	   | otherwise = False


{-
  5. (BONUS, 3p) Grafuri
  Se dă tipul Graph ce reprezintă un graf orientat sub forma unei liste de perechi
   în care primul element este un nod iar al doilea este lista de noduri spre
   care are arce;

   Definiți funcțiile:
   - graphToList, care convertește un Graph la o listă de arce, unde un arc
   este o pereche de valori (v1, v2) cu arcul plecând din v1
   - listToGraph, care face operația invers
                                           (Hint: pentru o abordare mai ușoară
   puteți implementa funcțiile helper addGraphNode (care adaugă un nod ”izolat”
   în graf) și addGraphEdge (care adauga un nod în lista de arce a altui nod)

-}

test5 :: TestPP ()
test5 = tests 5 3
        [ testVal (sort $ graphToList (Graph [ (1, [2, 3])
	                                     , (2, [4, 5])
					     , (3, [])
					     , (4, [5])
					     , (5, [])]))
	          graphList "graphToList" 1
        , testVal (sort $ graphToList $ listToGraph graphList) graphList "listToGraph" 2
	]
    where
          graphList = [(1, 2), (1, 3), (2, 4), (2, 5), (4, 5)]


data Graph a = Graph [(a, [a])] deriving Show

graphToList :: Graph a -> [(a, a)]
graphToList = concatMap (\(n, list) -> map ((,) n) list) . fromGraph
            where fromGraph (Graph x) = x


listToGraph :: Eq a => [(a, a)] -> Graph a
listToGraph = foldl integrate $ Graph []
            where integrate (Graph l) (s, d) = case lookup s l of
		                                        Nothing -> Graph $ (s, [d]) : l
				                        Just nn -> Graph $ (s, d : nn) : delete (s, nn) l

{-
Helpers for testing :)
-}
allTests = [test1, test2, test3, test4, test5]
check = runTestPP $ sequence_ allTests
[check1, check2, check3, check4, check5] = map runTestPP allTests
